# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the Error Analysis Manager class."""

from pathlib import Path
import json

from responsibleai.exceptions import UserConfigValidationException
from responsibleai._internal.constants import (
    ManagerNames, ListProperties, ErrorAnalysisManagerKeys as Keys)
from responsibleai._managers.base_manager import BaseManager
from responsibleai._config.base_config import BaseConfig
from responsibleai.exceptions import DuplicateManagerConfigException
from erroranalysis._internal.error_analyzer import ModelAnalyzer
from erroranalysis._internal.error_report import (
    json_converter as report_json_converter, as_error_report)
from responsibleai._interfaces import ErrorAnalysisData

REPORTS = 'reports'
CONFIG = 'config'
MAX_DEPTH = Keys.MAX_DEPTH
NUM_LEAVES = Keys.NUM_LEAVES
MIN_CHILD_SAMPLES = Keys.MIN_CHILD_SAMPLES
FILTER_FEATURES = Keys.FILTER_FEATURES
IS_COMPUTED = 'is_computed'


def config_json_converter(obj):
    """Helper function to convert ErrorAnalysisConfig object to json.

    :param obj: Object to convert to json.
    :type obj: object
    :return: The converted json.
    :rtype: dict
    """
    if isinstance(obj, ErrorAnalysisConfig):
        return obj.__dict__
    try:
        return obj.to_json()
    except AttributeError:
        return obj.__dict__


def as_error_config(json_dict):
    """Helper function to convert json to an ErrorAnalysisConfig object.

    :param json_dict: The json to convert.
    :type json_dict: dict
    :return: The converted ErrorAnalysisConfig.
    :rtype: ErrorAnalysisConfig
    """
    has_max_depth = MAX_DEPTH in json_dict
    has_num_leaves = NUM_LEAVES in json_dict
    has_min_child_samples = MIN_CHILD_SAMPLES in json_dict
    has_filter_features = FILTER_FEATURES in json_dict
    has_is_computed = IS_COMPUTED in json_dict
    has_all_fields = (has_max_depth and has_num_leaves and
                      has_min_child_samples and has_filter_features and
                      has_is_computed)
    if has_all_fields:
        max_depth = json_dict[MAX_DEPTH]
        num_leaves = json_dict[NUM_LEAVES]
        min_child_samples = json_dict[MIN_CHILD_SAMPLES]
        filter_features = json_dict[FILTER_FEATURES]
        config = ErrorAnalysisConfig(max_depth,
                                     num_leaves,
                                     min_child_samples,
                                     filter_features)
        config.is_computed = json_dict[IS_COMPUTED]
        return config
    else:
        return json_dict


class ErrorAnalysisConfig(BaseConfig):

    """Defines the ErrorAnalysisConfig, specifying the parameters to run.

    :param max_depth: The maximum depth of the tree.
    :type max_depth: int
    :param num_leaves: The number of leaves in the tree.
    :type num_leaves: int
    :param min_child_samples: The minimal number of data required to
        create one leaf.
    :type min_child_samples: int
    :param filter_features: One or two features to use for the
        matrix filter.
    :type filter_features: list
    """

    def __init__(self, max_depth, num_leaves,
                 min_child_samples, filter_features):
        """Defines the ErrorAnalysisConfig, specifying the parameters to run.

        :param max_depth: The maximum depth of the tree.
        :type max_depth: int
        :param num_leaves: The number of leaves in the tree.
        :type num_leaves: int
        :param min_child_samples: The minimal number of data required to
            create one leaf.
        :type min_child_samples: int
        :param filter_features: One or two features to use for the
            matrix filter.
        :type filter_features: list
        """
        super(ErrorAnalysisConfig, self).__init__()
        self.max_depth = max_depth
        self.num_leaves = num_leaves
        self.min_child_samples = min_child_samples
        self.filter_features = filter_features

    def __eq__(self, other_ea_config):
        """Returns true if this config is equal to the given config.

        :return: True if the given config is equal to the current config.
        :rtype: boolean
        """
        return (
            self.max_depth == other_ea_config.max_depth and
            self.num_leaves == other_ea_config.num_leaves and
            self.min_child_samples == other_ea_config.min_child_samples and
            self.filter_features == other_ea_config.filter_features
        )

    @property
    def __dict__(self):
        """Returns the dictionary representation of the ErrorAnalysisConfig.

        The dictionary contains the max depth, num leaves, min
        child samples and list of matrix filter features.

        :return: The dictionary representation of the ErrorAnalysisConfig.
        :rtype: dict
        """
        return {'max_depth': self.max_depth,
                'num_leaves': self.num_leaves,
                'min_child_samples': self.min_child_samples,
                'filter_features': self.filter_features,
                'is_computed': self.is_computed}

    def to_json(self):
        """Serialize ErrorAnalysisConfig object to json.

        :return: The string json representation of the ErrorAnalysisConfig.
        :rtype: str
        """
        return json.dumps(self, default=config_json_converter, indent=2)

    @staticmethod
    def from_json(json_str):
        """Deserialize json string to an ErrorAnalysisConfig object.

        :return: The deserialized ErrorAnalysisConfig.
        :rtype: ErrorAnalysisConfig
        """
        return json.loads(json_str, object_hook=as_error_config)


class ErrorAnalysisManager(BaseManager):

    """Defines the ErrorAnalysisManager for discovering errors in a model.

    :param model: The model to analyze errors on.
        A model that implements sklearn.predict or sklearn.predict_proba
        or function that accepts a 2d ndarray.
    :type model: object
    :param dataset: The dataset including the label column.
    :type dataset: pandas.DataFrame
    :param target_column: The name of the label column.
    :type target_column: str
    """

    def __init__(self, model, dataset, target_column,
                 categorical_features=None):
        """Defines the ErrorAnalysisManager for discovering errors in a model.

        :param model: The model to analyze errors on.
            A model that implements sklearn.predict or sklearn.predict_proba
            or function that accepts a 2d ndarray.
        :type model: object
        :param dataset: The dataset including the label column.
        :type dataset: pandas.DataFrame
        :param target_column: The name of the label column.
        :type target_column: str
        """
        self._true_y = dataset[target_column]
        self._dataset = dataset.drop(columns=[target_column])
        self._feature_names = list(self._dataset.columns)
        self._categorical_features = categorical_features
        self._ea_config_list = []
        self._ea_report_list = []
        self._analyzer = ModelAnalyzer(model,
                                       self._dataset,
                                       self._true_y,
                                       self._feature_names,
                                       self._categorical_features)

    def add(self, max_depth=3, num_leaves=31,
            min_child_samples=20, filter_features=None):
        """Add an error analyzer to be computed later.

        :param max_depth: The maximum depth of the tree.
        :type max_depth: int
        :param num_leaves: The number of leaves in the tree.
        :type num_leaves: int
        :param min_child_samples: The minimal number of data required to
            create one leaf.
        :type min_child_samples: int
        :param filter_features: One or two features to use for the
            matrix filter.
        :type filter_features: list
        """
        if self._analyzer.model is None:
            raise UserConfigValidationException(
                'Model is required for error analysis')

        ea_config = ErrorAnalysisConfig(
            max_depth=max_depth,
            num_leaves=num_leaves,
            min_child_samples=min_child_samples,
            filter_features=filter_features)
        is_duplicate = ea_config.is_duplicate(
            self._ea_config_list)

        if is_duplicate:
            raise DuplicateManagerConfigException(
                "Duplicate config specified for error analysis,"
                "config already added")
        else:
            self._ea_config_list.append(ea_config)

    def compute(self):
        """Creates an ErrorReport by running the error analyzer on the model.
        """
        for config in self._ea_config_list:
            if config.is_computed:
                continue
            config.is_computed = True
            max_depth = config.max_depth
            num_leaves = config.num_leaves
            min_child_samples = config.min_child_samples
            filter_features = config.filter_features
            report = self._analyzer.create_error_report(
                filter_features, max_depth=max_depth,
                min_child_samples=min_child_samples,
                num_leaves=num_leaves)
            self._ea_report_list.append(report)

    def get(self):
        """Get the computed error reports.

        Must be called after add and compute methods.

        :return: The computed error reports.
        :rtype: list[erroranalysis._internal.error_report.ErrorReport]
        """
        return self._ea_report_list

    def list(self):
        """List information about the ErrorAnalysisManager.

        :return: A dictionary of properties.
        :rtype: dict
        """
        props = {ListProperties.MANAGER_TYPE: self.name}
        reports = []
        for config in self._ea_config_list:
            report = {}
            report[Keys.IS_COMPUTED] = config.is_computed
            report[Keys.MAX_DEPTH] = config.max_depth
            report[Keys.NUM_LEAVES] = config.num_leaves
            report[Keys.MIN_CHILD_SAMPLES] = config.min_child_samples
            report[Keys.FILTER_FEATURES] = config.filter_features
            reports.append(report)
        props[Keys.REPORTS] = reports
        return props

    def get_data(self):
        """Get error analysis data

        :return: A array of ErrorAnalysisConfig.
        :rtype: List[ErrorAnalysisConfig]
        """
        return [
            self._get_error_analysis(i) for i in self.list()["reports"]]

    def _get_error_analysis(self, report):
        error_analysis = ErrorAnalysisData()
        error_analysis.maxDepth = report[Keys.MAX_DEPTH]
        error_analysis.numLeaves = report[Keys.NUM_LEAVES]
        error_analysis.minChildSamples = report[Keys.MIN_CHILD_SAMPLES]
        error_analysis.tree = self._analyzer.compute_error_tree(
            self._feature_names, None, None, error_analysis.maxDepth,
            error_analysis.numLeaves, error_analysis.minChildSamples)
        error_analysis.matrix = self._analyzer.compute_matrix(
            self._feature_names, None, None)
        return error_analysis

    @property
    def name(self):
        """Get the name of the error analysis manager.

        :return: The name of the error analysis manager.
        :rtype: str
        """
        return ManagerNames.ERROR_ANALYSIS

    def _save(self, path):
        """Save the ErrorAnalysisManager to the given path.

        :param path: The directory path to save the ErrorAnalysisManager to.
        :type path: str
        """
        top_dir = Path(path)
        top_dir.mkdir(parents=True, exist_ok=True)
        # save the reports
        reports_path = top_dir / REPORTS
        with open(reports_path, 'w') as file:
            json.dump(self._ea_report_list, file,
                      default=report_json_converter)
        # save the configs
        config_path = top_dir / CONFIG
        with open(config_path, 'w') as file:
            json.dump(self._ea_config_list, file,
                      default=config_json_converter)

    @staticmethod
    def _load(path, model_analysis):
        """Load the ErrorAnalysisManager from the given path.

        :param path: The directory path to load the ErrorAnalysisManager from.
        :type path: str
        :param model_analysis: The loaded parent ModelAnalysis.
        :type model_analysis: ModelAnalysis
        """
        # create the ErrorAnalysisManager without any properties using
        # the __new__ function, similar to pickle
        inst = ErrorAnalysisManager.__new__(ErrorAnalysisManager)
        top_dir = Path(path)
        reports_path = top_dir / REPORTS
        with open(reports_path, 'r') as file:
            ea_report_list = json.load(file, object_hook=as_error_report)
        inst.__dict__['_ea_report_list'] = ea_report_list
        config_path = top_dir / CONFIG
        with open(config_path, 'r') as file:
            ea_config_list = json.load(file, object_hook=as_error_config)
        inst.__dict__['_ea_config_list'] = ea_config_list
        categorical_features = model_analysis.categorical_features
        inst.__dict__['_categorical_features'] = categorical_features
        target_column = model_analysis.target_column
        true_y = model_analysis.test[target_column]
        dataset = model_analysis.test.drop(columns=[target_column])
        inst.__dict__['_dataset'] = dataset
        inst.__dict__['_true_y'] = true_y
        feature_names = list(dataset.columns)
        inst.__dict__['_feature_names'] = feature_names
        inst.__dict__['_analyzer'] = ModelAnalyzer(model_analysis.model,
                                                   dataset,
                                                   true_y,
                                                   feature_names,
                                                   categorical_features)
        return inst



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.
"""Result of causal analysis."""

import json
import pandas as pd
import re
import semver
import uuid

from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

from responsibleai._tools.causal.causal_constants import (
    ResultAttributes)
from responsibleai._tools.shared.versions import CausalVersions
from responsibleai._interfaces import (
    CausalConfig as CausalConfigInterface, CausalData, CausalPolicy,
    CausalPolicyGains, CausalPolicyTreeInternal, CausalPolicyTreeLeaf,
    ComparisonTypes)
from responsibleai._tools.causal.causal_config import CausalConfig
from responsibleai._tools.shared.base_result import BaseResult


class CausalResult(BaseResult['CausalResult']):
    """Result of causal analysis."""

    _ATTRIBUTES = [
        'config',
        'causal_analysis',
        'global_effects',
        'local_effects',
        'policies',
    ]

    def __init__(
        self,
        config: Optional[CausalConfig] = None,
        causal_analysis: Optional[Any] = None,
        global_effects: Optional[pd.DataFrame] = None,
        local_effects: Optional[pd.DataFrame] = None,
        policies: Optional[Any] = None,
    ):
        super().__init__(str(uuid.uuid4()), CausalVersions.get_current())

        self.config = config

        self.causal_analysis = causal_analysis
        self.global_effects = global_effects
        self.local_effects = local_effects
        self.policies = policies

    def _get_dashboard_object(self):
        causal_data = CausalData()

        causal_data.id = self.id
        causal_data.version = self.version

        causal_data.config = self._get_config_object(self.config)

        causal_data.global_effects = self.global_effects\
            .reset_index().to_dict(orient='records')
        local_dicts = self.local_effects.groupby('sample').apply(
            lambda x: x.reset_index().to_dict(
                orient='records')).values
        causal_data.local_effects = [list(v) for v in local_dicts]
        causal_data.policies = [self._get_policy_object(p) for p in
                                self.policies]

        return causal_data

    def _whatif(self, X, X_feature_new, feature_name, y, alpha=0.1):
        return self.causal_analysis.whatif(
            X, X_feature_new, feature_name, y, alpha=alpha)

    def _cohort_effects(self, X_test, alpha=0.01, keep_all_levels=False):
        return self.causal_analysis.cohort_causal_effect(
            X_test, alpha=alpha, keep_all_levels=keep_all_levels)

    def _get_config_object(self, config):
        config_object = CausalConfigInterface()
        config_object.treatment_features = config.treatment_features
        return config_object

    def _get_policy_object(self, policy):
        policy_object = CausalPolicy()
        policy_object.treatment_feature = policy[
            ResultAttributes.TREATMENT_FEATURE]
        policy_object.control_treatment = policy[
            ResultAttributes.CONTROL_TREATMENT]
        policy_object.local_policies = self._get_local_policies_object(
            policy[ResultAttributes.LOCAL_POLICIES])
        policy_object.policy_gains = self._get_policy_gains_object(
            policy[ResultAttributes.POLICY_GAINS])
        policy_object.policy_tree = self._get_policy_tree_object(
            policy[ResultAttributes.POLICY_TREE])
        return policy_object

    def _get_local_policies_object(self, local_policies):
        if local_policies is None:
            return None

        return local_policies.reset_index().\
            to_dict(orient='records')

    def _get_policy_gains_object(self, policy_gains):
        policy_gains_object = CausalPolicyGains()
        policy_gains_object.recommended_policy_gains = \
            policy_gains[ResultAttributes.RECOMMENDED_POLICY_GAINS]
        policy_gains_object.treatment_gains = \
            policy_gains[ResultAttributes.TREATMENT_GAINS]
        return policy_gains_object

    def _get_policy_tree_object(self, policy_tree):
        if policy_tree[ResultAttributes.LEAF]:
            policy_tree_object = CausalPolicyTreeLeaf()
            policy_tree_object.leaf = policy_tree[ResultAttributes.LEAF]
            policy_tree_object.n_samples = policy_tree[
                ResultAttributes.N_SAMPLES]
            policy_tree_object.treatment = policy_tree[
                ResultAttributes.TREATMENT]
        else:
            policy_tree_object = CausalPolicyTreeInternal()
            policy_tree_object.leaf = policy_tree[ResultAttributes.LEAF]
            feature, comparison, value = self._parse_comparison(
                policy_tree, self.config.categorical_features)
            policy_tree_object.feature = feature
            policy_tree_object.right_comparison = comparison
            policy_tree_object.comparison_value = value
            policy_tree_object.left = self._get_policy_tree_object(
                policy_tree[ResultAttributes.LEFT])
            policy_tree_object.right = self._get_policy_tree_object(
                policy_tree[ResultAttributes.RIGHT])
        return policy_tree_object

    @classmethod
    def _parse_comparison(
        cls,
        policy_tree: Any,
        categoricals: List[str],
    ) -> Tuple[str, str, Union[float, int, str]]:
        """Attempt to parse a categorical comparison from a policy tree node.

        The default assumption is that the feature is continuous and will
        have a real-valued threshold.

        This function checks known categorical features and attempts to parse
        a (feature, category) pair that would come from a one-hot encoding.

        Example (continuous):
            Original feature: "Size"
            Original comparison value: 0.891
            Parsed feature: "Size"
            Parsed comparison: "gt"
            Parsed comparison value: 0.891

        Example (categorical):
            Original feature: "Fruit_apple"
            Original comparison value: 0.5
            Parsed feature: "Fruit"
            Parsed comparison: "eq"
            Parsed comparison value: "apple"

        """
        tree_feature = policy_tree[ResultAttributes.FEATURE]
        threshold = policy_tree[ResultAttributes.THRESHOLD]

        # Default assumption is a continuous feature
        feature = tree_feature
        comparison = ComparisonTypes.GT
        value = threshold

        # Check for a categorical feature
        for cat_feature in sorted(categoricals)[::-1]:
            pattern = f'({cat_feature})_(.+)'
            match = re.match(pattern, tree_feature)
            if match is not None:
                feature = match.group(1)
                value = match.group(2)
                comparison = ComparisonTypes.EQ
                break

        return feature, comparison, value

    @classmethod
    def _get_schema(cls, version: str):
        cls._validate_version(version)

        schema_directory = Path(__file__).parent / 'dashboard_schemas'
        schema_filename = f'schema_{version}.json'
        schema_filepath = schema_directory / schema_filename
        with open(schema_filepath, 'r') as f:
            return json.load(f)

    @classmethod
    def _validate_version(cls, version):
        semver_version = semver.VersionInfo.parse(version)
        if semver_version.compare(CausalVersions.get_current()) > 0:
            raise ValueError("The installed version of responsibleai "
                             "is not compatible with causal result version "
                             f"{version}. Please upgrade in order to load "
                             "this result.")

        if version not in CausalVersions.get_all():
            raise ValueError(f"Invalid version for causal result: {version}")

    def _get_attributes(self):
        return self._ATTRIBUTES



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import json
import pandas as pd
import pytest

from jsonschema import ValidationError

from responsibleai import ModelTask
from responsibleai._managers.causal_manager import CausalManager
from responsibleai._tools.causal.causal_result import CausalResult
from responsibleai._tools.shared.versions import CausalVersions


@pytest.fixture(scope='module')
def causal_result(parks_data):
    train_df, test_df, target_feature = parks_data
    manager = CausalManager(train_df, test_df, target_feature,
                            ModelTask.REGRESSION, ['state', 'attraction'])

    return manager.add(['attraction'],
                       skip_cat_limit_checks=True,
                       upper_bound_on_cat_expansion=50)


class TestCausalVersioning:
    def test_current_roundtrip(self, tmpdir, causal_result):
        save_dir = tmpdir.mkdir('result-dir')
        result = causal_result
        result.save(save_dir)
        loaded = CausalResult.load(save_dir)
        assert loaded.id == result.id
        assert loaded.version == result.version
        assert loaded.version == CausalVersions.get_current()
        assert loaded._get_dashboard_data() == result._get_dashboard_data()
        assert loaded.global_effects.equals(result.global_effects)
        assert loaded.local_effects.equals(result.local_effects)
        self._check_recursive_df_equality(loaded.policies, result.policies)

    def _check_recursive_df_equality(self, o_1, o_2):
        assert type(o_1) == type(o_2)

        if isinstance(o_1, list):
            assert len(o_1) == len(o_2)
            for item_1, item_2 in zip(o_1, o_2):
                self._check_recursive_df_equality(item_1, item_2)
        elif isinstance(o_1, dict):
            assert o_1.keys() == o_2.keys()
            for key, item_1 in o_1.items():
                item_2 = o_2[key]
                self._check_recursive_df_equality(item_1, item_2)
        elif isinstance(o_1, pd.DataFrame):
            assert o_1.equals(o_2)
        else:
            assert o_1 == o_2

    def test_invalid_version_load(self, tmpdir, causal_result):
        save_dir = tmpdir.mkdir('result-dir')
        causal_result.save(save_dir)

        version_filepath = save_dir / 'version.json'
        with open(version_filepath, 'w') as f:
            new_version_data = {'version': 'invalid_version'}
            json.dump(new_version_data, f)

        message = r"invalid_version is not valid SemVer string"
        with pytest.raises(ValueError, match=message):
            CausalResult.load(save_dir)

    def test_invalid_dashboard_load(self, tmpdir, causal_result):
        save_dir = tmpdir.mkdir('result-dir')
        causal_result.save(save_dir)

        dashboard_filepath = save_dir / 'dashboard.json'
        with open(dashboard_filepath, 'w') as f:
            new_dashboard_data = {'fake_dashboard': 'invalid'}
            json.dump(new_dashboard_data, f)

        message = r"Failed validating"
        with pytest.raises(ValidationError, match=message):
            CausalResult.load(save_dir)

    def test_forward_incompatible(self, tmpdir, causal_result):
        save_dir = tmpdir.mkdir('result-dir')
        causal_result.save(save_dir)

        version_filepath = save_dir / 'version.json'
        with open(version_filepath, 'w') as f:
            new_version_data = {'version': '1000.0.0'}
            json.dump(new_version_data, f)

        message = ("The installed version of responsibleai is not "
                   "compatible with causal result version 1000.0.0. "
                   "Please upgrade in order to load this result.")
        with pytest.raises(ValueError, match=message):
            CausalResult.load(save_dir)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import os
import setuptools

# Version will be read from version.py
version = ''
name = 'responsibleai'
# Fetch Version
with open(os.path.join(name, '__version__.py')) as f:
    code = compile(f.read(), f.name, 'exec')
    exec(code)

# Fetch ReadMe
with open('README.md', 'r') as fh:
    long_description = fh.read()

# Use requirements.txt to set the install_requires
with open('requirements.txt') as f:
    install_requires = [line.strip() for line in f]

setuptools.setup(
    name=name,  # noqa: F821
    version=version,  # noqa: F821
    author="Roman Lutz, Ilya Matiach, Ke Xu",
    author_email="raiwidgets-maintain@microsoft.com",
    description="SDK API to explain models, and analyze "
                "errors in Machine Learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/microsoft/responsible-ai-widgets",
    packages=setuptools.find_packages(),
    package_data={
        '': [
            'schema_*.json',
        ]
    },
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha"
    ]
)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from responsibleai._tools.causal.causal_constants import DefaultParams


class TestCausalConstants:
    def test_tree_depth_limit(self):
        # The causal dashboard requires that this constant be no
        # greater than 2 in order to correctly display the
        # policy tree chart
        assert DefaultParams.DEFAULT_MAX_TREE_DEPTH == 2



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import pytest
import re
import semver

from responsibleai._tools.shared.versions import CausalVersions


class TestVersions:

    @pytest.mark.parametrize(
        'versions_class',
        [
            CausalVersions,
        ]
    )
    def test_all_versions_valid(self, versions_class):
        version_strings = versions_class.get_all()
        assert len(version_strings) > 0

        current = semver.VersionInfo.parse(versions_class.get_current())
        for version_string in version_strings:
            # Check valid semantic version
            version = semver.VersionInfo.parse(version_string)

            # Check current version is the latest version
            assert version.compare(current) <= 0

    @pytest.mark.parametrize(
        'versions_class',
        [
            CausalVersions,
        ]
    )
    def test_all_versions_unique(self, versions_class):
        versions = versions_class.get_all()
        assert len(versions) > 0
        s = set()
        for version in versions:
            print(version, s)
            assert version not in s, f"Version {version} is duplicated"
            s.add(version)

    @pytest.mark.parametrize(
        'versions_class',
        [
            CausalVersions,
        ]
    )
    def test_version_name_formats(self, versions_class):
        for name, _ in vars(versions_class).items():
            if name.startswith('__'):
                continue
            if name != name.upper():
                continue
            pattern = r'^V_\d+_\d+_\d+$'
            match = re.match(pattern, name)
            assert match is not None, name



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.
"""BaseResult of ResponsibleAI tools."""

import json
import jsonschema

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Generic, List, TypeVar, Union

from responsibleai.serialization_utilities import serialize_json_safe
from responsibleai._tools.shared.attribute_serialization import (
    load_attributes, save_attributes)
from responsibleai._internal.constants import (
    SerializationAttributes)


TResult = TypeVar('TResult')  # Type for subclasses of BaseResult


class BaseResult(ABC, Generic[TResult]):
    def __init__(self, id: str, version: str) -> TResult:
        """Construct a BaseResult.

        :param id: ID of the result.
        :param version: Version of the result.
        :return: New BaseResult.
        """
        self.id = id
        self.version = version

        self._dashboard_data = None

    def save(self, path: Union[str, Path]) -> None:
        """Save the result to disk at the provided path.

        :param path: Path to result directory on disk.
        """
        result_dir = Path(path)
        result_dir.mkdir(parents=True, exist_ok=True)

        # Save metadata like ID and and version to disk
        self._save_metadata(result_dir)

        # Save all result attributes to disk
        save_attributes(self, self._get_attributes(), result_dir)

        # Save dashboard data
        self._save_dashboard(result_dir)

    def _save_metadata(self, result_dir: Path):
        """Save result metadata to disk.

        :param path: Path to result directory on disk.
        """
        id_path = result_dir / SerializationAttributes.ID_FILENAME
        with open(id_path, 'w') as f:
            json.dump({SerializationAttributes.ID_KEY: self.id}, f)

        version_path = result_dir / SerializationAttributes.VERSION_FILENAME
        with open(version_path, 'w') as f:
            json.dump({SerializationAttributes.VERSION_KEY: self.version}, f)

    def _save_dashboard(self, result_dir: Path):
        """Save result dashboard to disk.

        :param path: Path to result directory on disk.
        """
        dashboard = self._get_dashboard_data()
        dashboard_filename = SerializationAttributes.DASHBOARD_FILENAME
        with open(result_dir / dashboard_filename, 'w') as f:
            json.dump(dashboard, f)

    @classmethod
    def load(cls, path: Union[str, Path]) -> TResult:
        """Load a result from disk at the provided path.

        :param path: Path to result directory on disk.
        """
        result_dir = Path(path)
        loaded = cls()

        # Load metadata like ID and version
        loaded._load_metadata(result_dir)

        # Load all result attributes from disk
        load_attributes(loaded, loaded._get_attributes(), result_dir)

        # Load dashboard data
        loaded._load_dashboard(result_dir)

        loaded._validate_schema()
        return loaded

    def _load_metadata(self, result_dir: Path) -> str:
        """Load result metadata from disk.

        :param path: Path to result directory on disk.
        """
        id_path = result_dir / SerializationAttributes.ID_FILENAME
        with open(id_path, 'r') as f:
            self.id = json.load(f)[SerializationAttributes.ID_KEY]

        version_path = result_dir / SerializationAttributes.VERSION_FILENAME
        with open(version_path, 'r') as f:
            self.version = json.load(f)[SerializationAttributes.VERSION_KEY]

    def _load_dashboard(self, result_dir: Path) -> str:
        """Load result dashboard from disk.

        :param path: Path to result directory on disk.
        """
        dashboard_filename = SerializationAttributes.DASHBOARD_FILENAME
        with open(result_dir / dashboard_filename, 'r') as f:
            self._dashboard_data = json.load(f)

    def _validate_schema(self):
        schema = self._get_schema(self.version)
        jsonschema.validate(self._get_dashboard_data(), schema)

    @abstractmethod
    def _get_dashboard_object(self):
        """Get the dashboard object for the result.

        This object is created with the shared interfaces used by the
        ResponsibleAI widgets to ensure consistent contracts.

        See responsibleai/_interfaces.py for existing interfaces.
        """
        pass

    def _get_dashboard_data(self):
        """Get the Python dict representation of the dashboard object."""
        if self._dashboard_data is None:
            dashboard_object = self._get_dashboard_object()
            self._dashboard_data = serialize_json_safe(dashboard_object)

        return self._dashboard_data

    @classmethod
    @abstractmethod
    def _get_schema(cls, version: str) -> Any:
        pass

    @abstractmethod
    def _get_attributes(self) -> List[str]:
        pass



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.
import pytest

from erroranalysis._internal.matrix_filter import (
    CATEGORY1, CATEGORY2, COUNT, FALSE_COUNT, MATRIX, VALUES)
from responsibleai.exceptions import DuplicateManagerConfigException, \
    UserConfigValidationException
from responsibleai._internal.constants import (
    ErrorAnalysisManagerKeys as Keys)

SIZE = 'size'
PARENTID = 'parentId'
ERROR = 'error'
ID = 'id'
MIN_CHILD_SAMPLES = Keys.MIN_CHILD_SAMPLES
REPORTS = Keys.REPORTS
FILTER_FEATURES = Keys.FILTER_FEATURES


def setup_error_analysis(model_analysis, add_ea=True, max_depth=3):
    if add_ea:
        if model_analysis.model is None:
            with pytest.raises(UserConfigValidationException,
                               match='Model is required for error analysis'):
                model_analysis.error_analysis.add(max_depth=max_depth)
        else:
            model_analysis.error_analysis.add(max_depth=max_depth)
            with pytest.raises(DuplicateManagerConfigException):
                model_analysis.error_analysis.add(max_depth=max_depth)
    model_analysis.error_analysis.compute()


def validate_error_analysis(model_analysis, expected_reports=1):
    if model_analysis.model is None:
        return
    reports = model_analysis.error_analysis.get()
    assert isinstance(reports, list)
    assert len(reports) == expected_reports
    ea_info_list = model_analysis.error_analysis.list()
    for idx, report in enumerate(reports):
        matrix = report.matrix
        matrix_features = report.matrix_features
        tree_features = report.tree_features
        info_features = ea_info_list[REPORTS][idx][FILTER_FEATURES]
        assert matrix_features == info_features
        assert tree_features == model_analysis.error_analysis._feature_names

        ea_x_dataset = model_analysis.error_analysis._dataset
        ea_true_y = model_analysis.error_analysis._true_y

        expected_count = len(ea_x_dataset)
        if matrix is not None:
            predictions = model_analysis.model.predict(ea_x_dataset)
            expected_false_count = sum(predictions != ea_true_y)
            validate_matrix(matrix, expected_count, expected_false_count)

        tree = report.tree
        min_child_samples = ea_info_list[REPORTS][idx][MIN_CHILD_SAMPLES]
        validate_tree(tree, expected_count, min_child_samples)


def validate_matrix(matrix, exp_total_count, exp_total_false_count):
    assert MATRIX in matrix
    assert CATEGORY1 in matrix
    assert CATEGORY2 in matrix
    num_cat1 = len(matrix[CATEGORY1][VALUES])
    num_cat2 = len(matrix[CATEGORY2][VALUES])
    assert len(matrix[MATRIX]) == num_cat1
    assert len(matrix[MATRIX][0]) == num_cat2
    # take sum of count, false count
    total_count = 0
    total_false_count = 0
    for i in range(num_cat1):
        for j in range(num_cat2):
            total_count += matrix[MATRIX][i][j][COUNT]
            total_false_count += matrix[MATRIX][i][j][FALSE_COUNT]
    assert exp_total_count == total_count
    assert exp_total_false_count == total_false_count


def validate_tree(tree, expected_count, min_child_samples):
    assert tree is not None
    assert len(tree) > 0
    assert ERROR in tree[0]
    assert ID in tree[0]
    assert PARENTID in tree[0]
    assert tree[0][PARENTID] is None
    assert SIZE in tree[0]
    assert tree[0][SIZE] == expected_count
    for node in tree:
        assert node[SIZE] >= min_child_samples



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.
"""Utilities for attribute serialization."""

import json
import pickle

from pathlib import Path
from typing import Any, List, Union


class SerializationFormats:
    PICKLE = 'pickle'
    JSON = 'json'
    TEXT = 'text'


class SerializationExtensions:
    PKL = 'pkl'
    JSON = 'json'
    TXT = 'txt'

    @classmethod
    def from_format(cls, file_format: str) -> str:
        if file_format == SerializationFormats.PICKLE:
            return cls.PKL
        elif file_format == SerializationFormats.JSON:
            return cls.JSON
        elif file_format == SerializationFormats.TEXT:
            return cls.TXT
        else:
            raise ValueError(f"Unknown format: {file_format}")


def save_attributes(
    o: Any,
    attributes: List[str],
    path: Union[str, Path],
    file_format: Union[str, List[str]] = SerializationFormats.PICKLE,
) -> List[Path]:
    """Save attributes from an object to disk.

    :param o: Object from which to pull attributes.
    :param attributes: List of attributes on the object to save.
    :param path: Path to directory on disk in which to write the attributes.
    :param file_format: File format to use when writing to disk. A list of
        file formats can be passed to assign each attribute a different
        format.
    :returns: List of paths to the saved attributes.
    """
    paths = []
    dir_path = Path(path)
    is_format_list = isinstance(file_format, list)
    for i, attribute in enumerate(attributes):
        attribute_format = file_format[i] if is_format_list else file_format
        value = getattr(o, attribute)
        extension = SerializationExtensions.from_format(attribute_format)
        path = dir_path / f'{attribute}.{extension}'
        _save_attribute(value, path, attribute_format)
        paths.append(path)
    return paths


def _save_attribute(
    value: Any,
    path: Union[str, Path],
    file_format: str,
) -> None:
    if file_format == SerializationFormats.PICKLE:
        with open(path, 'wb') as f:
            pickle.dump(value, f)
    elif file_format == SerializationFormats.JSON:
        with open(path, 'w') as f:
            json.dump(value, f)
    elif file_format == SerializationFormats.TEXT:
        with open(path, 'w') as f:
            f.write(value)
    else:
        raise ValueError(f"Invalid format {file_format}")


def load_attributes(
    o: Any,
    attributes: List[str],
    path: Union[str, Path],
    file_format: Union[str, List[str]] = SerializationFormats.PICKLE,
    fail_on_missing: bool = True,
) -> None:
    """Load attributes from disk and save to an existing object.

    :param o: Object on which to save the loaded attributes.
    :param attributes: List of attributes to load to the object.
    :param path: Path to directory on disk where attributes are saved.
    :param file_format: File format to use when loading attributes from
        disk. A list of file formats can be passed to assign each
        attribute a different format.
    :param fail_on_missing: Whether to raise an exception if an attribute
        was not found.
    """
    dir_path = Path(path)
    is_format_list = isinstance(file_format, list)
    for i, attribute in enumerate(attributes):
        attribute_format = file_format[i] if is_format_list else file_format
        extension = SerializationExtensions.from_format(attribute_format)
        path = dir_path / f'{attribute}.{extension}'
        if not fail_on_missing and (not path.exists() or not path.is_file()):
            continue
        value = _load_attribute(path, attribute_format)
        setattr(o, attribute, value)


def _load_attribute(
    path: Union[str, Path],
    file_format: str,
) -> Any:
    if file_format == SerializationFormats.PICKLE:
        with open(path, 'rb') as f:
            return pickle.load(f)
    elif file_format == SerializationFormats.JSON:
        with open(path, 'r') as f:
            return json.load(f)
    elif file_format == SerializationFormats.TEXT:
        with open(path, 'r') as f:
            return f.read()
    else:
        raise ValueError(f"Invalid format {file_format}")
